import numpy as np
import random
import pickle
import argparse
import time
from solver import solve_ilp_opt, solve_ilp_sat
from mk_problem import mk_batch_problem


def gen_instance_arrays(args, n, dtype):
    weight = []
    profit = []
    for _ in range(n):
        wi = random.randint(1, args.R)
        # if wi % 2 == 1: wi += 1   # assert wi is even
        weight.append(wi)

    assert(dtype in [1, 2, 3, 4])
    for i in range(n):
        if dtype == 1:
            # uncorr.
            profit.append(random.randint(1, args.R))
        elif dtype == 2:
            # weakly corr.
            rr = int(args.R / 10)
            pf = weight[i] - rr + random.randint(0, 2 * rr)
            if pf <= 0: pf = 1
            profit.append(pf)
        elif dtype == 3:
            # strongly corr.
            profit.append(weight[i] + 10)
        else:
            # subset sum
            profit.append(weight[i])
    return weight, profit


def gen_iconstraints(args, n):
    dtype = args.type
    if dtype == 5:
        dtype = random.randint(1, 4)
    weight, profit = gen_instance_arrays(args, n, dtype)
    c_limit = random.randint(1, sum(weight))
    # c_limit = int(sum(weight) / 2)
    # if c_limit % 2 == 0: c_limit += 1    # assert c_limit is odd
    best_profit = solve_ilp_opt(n, weight, profit, c_limit)
    return n, weight, profit, c_limit, best_profit


def normalization(iconstraints_left, iconstraints_right):
    for iconstraint_id, iconstraint_left in enumerate(iconstraints_left):
        right_value = iconstraints_right[iconstraint_id]
        for var, ratio in iconstraint_left:
            if ratio < 0:  # -cx => -c + c(~x)
                right_value -= ratio
        iconstraints_right[iconstraint_id] = right_value
    iconstr_norm = []
    for i in range(len(iconstraints_left)):
        if iconstraints_right[i] <= 0:  # if b <= 0, constraint is trivial SAT
            continue
        else:
            icon = []
            for var, ratio in iconstraints_left[i]:
                icon.append(tuple([var, 1.0 * ratio / iconstraints_right[i]]))
            iconstr_norm.append(icon)  # icon >= 1
    return iconstr_norm


def test_cplex_time(n_vars, weight, profit, c_limit, right_val, is_sat_guess):
    icon_1 = []
    icon_2 = []
    for i in range(n_vars):
        icon_1.append(tuple([i, -weight[i]]))
        icon_2.append(tuple([i, profit[i]]))
    iconstraints_left = [icon_1, icon_2]
    iconstraints_right = [-c_limit, right_val]
    time_st = time.perf_counter()
    is_sat = solve_ilp_sat(n_vars, iconstraints_left, iconstraints_right)
    time_ed = time.perf_counter()
    if is_sat != is_sat_guess:
        print("[WARNING] Contradictory decision")
    # print("gentime", time_ed - time_st)
    return time_ed - time_st


def gen_instance(args, n):
    while True:
        n_vars, weight, profit, c_limit, best_profit = gen_iconstraints(args, n)
        rr = int(args.R / 10)
        right_val = best_profit - rr + random.randint(0, 2 * rr)  # right_val = best_profit ± rr
        if right_val <= 0: right_val = 1

        is_sat = None
        if right_val <= best_profit: is_sat = True
        else: is_sat = False

        icon_1 = []
        icon_2 = []
        for i in range(n_vars):
            icon_1.append(tuple([i, -weight[i]]))
            icon_2.append(tuple([i, profit[i]]))
        iconstraints_left = [icon_1, icon_2]
        iconstraints_right = [-c_limit, right_val]
        iconstr_norm = normalization(iconstraints_left, iconstraints_right)
        if len(iconstr_norm) == 2: break  # not trivial, return
    # cplex_time = test_cplex_time(n_vars, weight, profit, c_limit, right_val, is_sat)
    # return n_vars, iconstr_left, iconstr_right, is_sat, cplex_time
    return n_vars, iconstr_norm, is_sat


def generate(args):

    f = open(args.gen_log, 'w')
    n_cnt = args.max_n - args.min_n + 1
    problems_per_n = args.n_pairs * 1.0 / n_cnt

    problems = []
    batches = []
    n_nodes_in_batch = 0
    prev_n_vars = args.min_n
    # tot_gen_time = 0
    # tot_gen_cnt = 0

    for n_var in range(args.min_n, args.max_n + 1):
        lower_bound = int((n_var - args.min_n) * problems_per_n)
        upper_bound = int((n_var - args.min_n + 1) * problems_per_n)
        for problems_idx in range(lower_bound, upper_bound):
            n_vars, iconstr_norm, is_sat = gen_instance(args, n_var)

            n_constraints = len(iconstr_norm)
            n_nodes = 2 * n_vars + n_constraints
            if n_nodes > args.max_nodes_per_batch:
                continue

            batch_ready = False
            if args.one == 1:
                batch_ready = True
            if prev_n_vars and n_vars != prev_n_vars:
                batch_ready = True
            elif n_nodes_in_batch + n_nodes > args.max_nodes_per_batch:
                batch_ready = True

            if batch_ready:
                batches.append(mk_batch_problem(problems))
                print("  batch %d done (%d vars, %d problems)" %
                      (len(batches), prev_n_vars, len(problems)))
                print("batch %d done (%d vars, %d problems)" %
                      (len(batches), prev_n_vars, len(problems)), file=f, flush=True)
                del problems[:]
                n_nodes_in_batch = 0

            prev_n_vars = n_vars
            # is_sat = solve_ilp_sat(n_vars, iconstraints_left, iconstraints_right)
            dstr = "prob_n=%d_C=%d_R=%d_tp=%d_id=%d_sat=%d" % (n_vars, args.C, args.R, args.type, problems_idx, is_sat)
            problems.append((dstr, n_vars, iconstr_norm, is_sat))
            print("  problem generation done (%d vars, %d constraints, %s)" %
                  (n_vars, len(iconstr_norm), 'SAT' if is_sat else 'UNSAT'), file=f, flush=True)
            n_nodes_in_batch += n_nodes
    # print("sumgentime", tot_gen_time / tot_gen_cnt)
    if len(problems) > 0:
        batches.append(mk_batch_problem(problems))
        print("  batch %d done (%d vars, %d problems)" % (len(batches), n_vars, len(problems)))
        print("batch %d done (%d vars, %d problems)" % (len(batches), n_vars, len(problems)), file=f, flush=True)
        del problems[:]
    return batches


if __name__ == "__main__":

    print("Start Data Generation...")

    parser = argparse.ArgumentParser()
    parser.add_argument('out_dir', action='store', type=str)
    parser.add_argument('gen_log', action='store', type=str)
    parser.add_argument('n_pairs', action='store', type=int)
    parser.add_argument('max_nodes_per_batch', action='store', type=int)

    parser.add_argument('--min_n', action='store', dest='min_n', type=int, default=40)
    parser.add_argument('--max_n', action='store', dest='max_n', type=int, default=40)
    parser.add_argument('--one', action='store', dest='one', type=int, default=0)

    parser.add_argument('--type', action='store', dest='type', type=int, default=1)
    parser.add_argument('--C', action='store', dest='C', type=int, default=100)
    parser.add_argument('--R', action='store', dest='R', type=int, default=100)

    parser.add_argument('--py_seed', action='store', dest='py_seed', type=int, default=None)
    parser.add_argument('--np_seed', action='store', dest='np_seed', type=int, default=None)

    args = parser.parse_args()

    if args.py_seed is not None:
        random.seed(args.py_seed)
    if args.np_seed is not None:
        np.random.seed(args.np_seed)

    batches = generate(args)

    # create directory
    # if not os.path.exists(args.out_dir):
    #   os.mkdir(args.out_dir)

    dataset_filename = "%s/kp_v%dto%d_tp%d_npb%d_nb%d.pkl" % (args.out_dir,
                                                              args.min_n,
                                                              args.max_n,
                                                              args.type,
                                                              args.max_nodes_per_batch,
                                                              len(batches))
    print("Writing %d batches to %s..." % (len(batches), dataset_filename))
    with open(dataset_filename, 'wb') as f_dump:
        pickle.dump(batches, f_dump)
