import pickle
import os
import random
from tqdm import tqdm

import torch
import torch.nn as nn
import torch.optim as optim

from neuroilp import NeuroILP
from config import parser


def adjust_learning_rate(optimizer, epoch, init_lr):
    if epoch >= 400:
        lr = init_lr / 100
    elif epoch >= 200:
        lr = init_lr / 10
    optimizer.param_groups[0]['lr'] = lr


def find_train_files(args):
	train_fnames = dict()
	fnames = os.listdir(os.path.join(args.data_dir, 'train'))
	for n_var in range(args.min_n, args.max_n + 1):
		for fname in fnames:
			if ("kp_v%dto%d_" % (n_var, n_var)) in fname:
				train_fnames[n_var] = fname
				break
	return train_fnames


def sample_train_data(args, train_fnames):
    n_cnt = args.max_n - args.min_n + 1
    problems_per_n = args.n_pairs / n_cnt
    train = list()
    for n_var in range(args.min_n, args.max_n + 1):
        with open(os.path.join(args.data_dir, 'train', train_fnames[n_var]), 'rb') as f:
            batches = pickle.load(f)
            n_problem_cnt = 0
            while n_problem_cnt < problems_per_n:
                batch = random.choice(batches)
                train.append(batch)
                n_problem_cnt += len(batch.is_sat)
    return train


args = parser.parse_args()
torch.manual_seed(args.py_seed)

net = NeuroILP(args)
net = net.cuda()

task_name = args.task_name + \
            '_sr' + str(args.min_n) + 'to' + str(args.max_n) + \
            '_ep' + str(args.epochs) + \
            '_nr' + str(args.n_rounds) + \
            '_d' + str(args.dim)
simple_log_file = open(os.path.join(args.log_dir, task_name + '.log'), 'a+')
detail_log_file = open(os.path.join(args.log_dir, task_name + '_details.log'), 'a+')

val = None
with open(os.path.join(args.data_dir, 'val', args.val_file), 'rb') as f:
    val = pickle.load(f)
print('Num of validate batches:', len(val), file=simple_log_file, flush=True)

train_fnames = find_train_files(args)
print('Num of train files loaded:', len(train_fnames), file=simple_log_file, flush=True)

init_lr = 0.00008 * 2
loss_fn = nn.BCELoss()
optim = optim.Adam(net.parameters(), lr=init_lr, weight_decay=1e-10)
sigmoid = nn.Sigmoid()

best_acc = 0.0
start_epoch = 0

if args.restore is not None:
    print('Restore model from', args.restore)
    print('Restore model from', args.restore, file=simple_log_file, flush=True)
    model = torch.load(os.path.join(args.model_dir, args.restore))
    start_epoch = model['epoch']
    best_acc = model['acc']
    net.load_state_dict(model['state_dict'])

for epoch in range(start_epoch, args.epochs):
    train = sample_train_data(args, train_fnames)
    print('Num of sampled train batches:', len(train))
    print('Num of sampled train batches:', len(train), file=simple_log_file, flush=True)

    print('==> %d/%d epoch, prev best acc: %.3f' % (epoch + 1, args.epochs, best_acc))
    print('==> %d/%d epoch, prev best acc: %.3f' % (epoch + 1, args.epochs, best_acc), file=simple_log_file, flush=True)
    print('==> %d/%d epoch, perv best acc: %.3f' % (epoch + 1, args.epochs, best_acc), file=detail_log_file, flush=True)
    
    print("* Results on the training set", file=detail_log_file, flush=True)
    train_bar = tqdm(train)
    # lr = adjust_learning_rate(optim, epoch, init_lr)
    TP, TN, FN, FP = torch.zeros(1).long(), torch.zeros(1).long(), torch.zeros(1).long(), torch.zeros(1).long()
    net.train()
    for _, prob in enumerate(train_bar):
        optim.zero_grad()
        outputs = net(prob)
        outputs = sigmoid(outputs)
        target = torch.Tensor(prob.is_sat).cuda().float()
        loss = loss_fn(outputs, target)
        desc = 'loss: %.4f; ' % (loss.item())

        loss.backward()
        optim.step()

        preds = torch.where(outputs > 0.5, torch.ones(outputs.shape).cuda(), torch.zeros(outputs.shape).cuda())

        TP += (preds.eq(1) & target.eq(1)).cpu().sum()
        TN += (preds.eq(0) & target.eq(0)).cpu().sum()
        FN += (preds.eq(0) & target.eq(1)).cpu().sum()
        FP += (preds.eq(1) & target.eq(0)).cpu().sum()
        TOT = TP + TN + FN + FP

        desc += 'acc: %.3f, TP: %.3f, TN: %.3f, FN: %.3f, FP: %.3f' % (
            (TP.item() + TN.item()) * 1.0 / TOT.item(),
            TP.item() * 1.0 / TOT.item(),
            TN.item() * 1.0 / TOT.item(),
            FN.item() * 1.0 / TOT.item(),
            FP.item() * 1.0 / TOT.item())
        # train_bar.set_description(desc)
        if (_ + 1) % 100 == 0:
            print(desc, file=detail_log_file, flush=True)
    print(desc, file=detail_log_file, flush=True)
    print(desc, file=simple_log_file, flush=True)

    print("* Results on the validation set", file=detail_log_file, flush=True)
    val_bar = tqdm(val)
    TP, TN, FN, FP = torch.zeros(1).long(), torch.zeros(1).long(), torch.zeros(1).long(), torch.zeros(1).long()
    net.eval()
    for _, prob in enumerate(val_bar):
        optim.zero_grad()
        outputs = net(prob)
        outputs = sigmoid(outputs)
        # print(outputs.shape, target.shape)
        # print(outputs, target)
        target = torch.Tensor(prob.is_sat).cuda().float()
        preds = torch.where(outputs > 0.5, torch.ones(outputs.shape).cuda(), torch.zeros(outputs.shape).cuda())

        TP += (preds.eq(1) & target.eq(1)).cpu().sum()
        TN += (preds.eq(0) & target.eq(0)).cpu().sum()
        FN += (preds.eq(0) & target.eq(1)).cpu().sum()
        FP += (preds.eq(1) & target.eq(0)).cpu().sum()
        TOT = TP + TN + FN + FP

        desc = 'acc: %.3f, TP: %.3f, TN: %.3f, FN: %.3f, FP: %.3f' % ((
            TP.item() + TN.item()) * 1.0 / TOT.item(),
            TP.item() * 1.0 / TOT.item(),
            TN.item() * 1.0 / TOT.item(),
            FN.item() * 1.0 / TOT.item(),
            FP.item() * 1.0 / TOT.item())
        # val_bar.set_description(desc)
        if (_ + 1) % 50 == 0:
            print(desc, file=detail_log_file, flush=True)
    print(desc, file=detail_log_file, flush=True)
    print(desc, file=simple_log_file, flush=True)

    acc = (TP.item() + TN.item()) * 1.0 / TOT.item()
    torch.save({'epoch': epoch + 1, 'acc': acc, 'state_dict': net.state_dict()},
               os.path.join(args.model_dir, task_name + '_last.pth.tar'))
    if acc >= best_acc:
        best_acc = acc
        torch.save({'epoch': epoch + 1, 'acc': best_acc, 'state_dict': net.state_dict()},
                   os.path.join(args.model_dir, task_name + '_best.pth.tar'))
